#include <algorithm>
#include <iostream>
#include <vector>
#include <utility>

#include "Util.h"

using namespace std;

template<typename T>
void BubbleSort(vector<T>& v) {
  for (size_t i = v.size() - 1; i > 0; --i) {
    bool swapped = false;
    for (size_t j = 0; j < i; ++j) {
      if (v[j] > v[j + 1]) {
        swap(v[j], v[j + 1]);
        swapped = true;
      }
    }
    if (!swapped) {
      // the list is already sorted
      return;
    }
  }
}

template<typename T>
void SelectionSort(vector<T>& v) {
  // Idea: go through the array, find the largest element and put it to the last
  // slot, then rinse & repeat for the rest of the array
  for (size_t i = v.size() - 1; i > 0; --i) {
    size_t largest_index = 0;
    for (size_t j = 1; j <= i; ++j) {
      if (v[j] > v[largest_index]) {
        largest_index = j;
      }
    }
    if (largest_index != i) {
      swap(v[largest_index], v[i]);
    }
  }
}

template<typename T>
void InsertionSort(vector<T>& v) {
  // Idea: sort the prefix of the array, then "insert" the current element to
  // the appropriate slot in the prefix.


  // The correct position we want to find
  ssize_t where_to_put;
  for (size_t i = 1; i < v.size(); ++i) {
    where_to_put = i - 1;
    // move out the previous value because it will be overwritten
    T item = move(v[i]);
  
    // shift the array while looking for where to put the value
    while (where_to_put >= 0 && v[where_to_put] > item) {
      v[where_to_put + 1] = v[where_to_put];
      where_to_put -= 1;
    }
    // insert the value back
    v[where_to_put + 1] = move(item);
  
    // we performed "i - where_to_put + 1" moves/swaps, record it
    n_swaps_ += i - where_to_put + 1;
  }
}

// 1 4 3 2 6 5

template<typename T>
void Merge(const typename vector<T>::iterator begin,
           const typename vector<T>::iterator middle,
           const typename vector<T>::iterator end) {
  // a temporary vector that will contain the sorted vector, we will move the
  // elements back after merging.
  vector<T> tmp;
  tmp.reserve(end - begin);

  // 1, 2, 3, 4, 0, 5, 6, 7, end
  // ^ begin     ^ middle

  auto left = begin;
  auto right = middle;

  while (left != middle && right != end) {
    if (*left <= *right) {
      tmp.push_back(move(*left));
      left++;
    } else {
      tmp.push_back(move(*right));
      right++;
    }
  }

  while (left != middle) {
    tmp.push_back(move(*left));
    left++;
  }

  while (right != end) {
    tmp.push_back(move(*right));
    right++;
  }

  auto j = begin;

  for (auto i = tmp.begin(); i != tmp.end(); ++i, ++j) {
    *j = move(*i);
  }
}

// begin is inclusive, end is exclusive.
template<typename T>
void MergeSortImpl(typename vector<T>::iterator begin,
                   typename vector<T>::iterator end) {
  if (end - begin > 1) {
    auto middle = begin + (end - begin) / 2;

    MergeSortImpl<T>(begin, middle);
    MergeSortImpl<T>(middle, end);
    Merge<T>(begin, middle, end);
  }
}

template<typename T>
void MergeSort(vector<T>& v) {
  MergeSortImpl<T>(v.begin(), v.end());
}

template<typename Iterator>
Iterator MedianOf3(Iterator a, Iterator b, Iterator c) {
  if (*a <= *b) {
    if (*a < *c) {
      if (*b < *c) {
        return b;
      }
      return c;
    }
    return a;
  }
  return MedianOf3(b, a, c);
}

template<typename Iterator>
Iterator Partition(Iterator begin, Iterator end) {
  // Choose the first value as the pivot
  Iterator pivot = begin;
  // We will use two iterators to go through the array in opposite direction,
  // and find some elements out-of-order-relative-to-the-pivot.
  Iterator left = begin + 1;
  Iterator right = end - 1;

  // cout << "pivot: " << *pivot << " l: " << *left << " r: " << *right << '\n';

  // Example run for arr[2:end] (after establishing 1 as a pivot) in the
  // previous run.
  //
  // 0 1 4 2 3 3'6 7 5
  //     p     r l     

  while (left <= right) {
    // move right while *left <= *pivot
    while (left != end && *left <= *pivot) {
      ++left;
    } // left-> 6
    // move left while *right > *pivot
    while (right != begin /* unnecessary */ && *right >= *pivot) {
      --right;
    } // right-> 3'

    // swap *left and *right if they are in the wrong order
    if (left < right) {
      swap(*left, *right);
    }
  }

  // the final pivot should be at the index of `right`, put it in place
  swap(*pivot, *right);
  return right; // the location of the pivot
}

template<typename Iterator>
void QuicksortImpl(Iterator begin, Iterator end) {
  if (end - begin <= 1) {
    return;
  }

  Iterator pivot = Partition(begin, end);

  // sort the smaller array first, then the larger one
  if (pivot - begin > end - pivot - 1) {
    QuicksortImpl(pivot + 1, end);
    QuicksortImpl(begin, pivot);
  } else {
    QuicksortImpl(begin, pivot);
    QuicksortImpl(pivot + 1, end);
  }
}

template<typename T>
void Quicksort(vector<T>& v) {
  QuicksortImpl(v.begin(), v.end());
}

// Check if the given collection is sorted
template<typename Collection>
bool IsSorted(const Collection& v) {
  for (auto i = v.begin() + 1; i != v.end(); ++i) {
    if (*(i - 1) > *i) {
      return false;
    }
  }
  return true;
}
  

// Copy the vector, and run the algorithm & report stats
template<typename T = FancyInt>
void RunAlgorithm(vector<T> v,
                  const std::string & name,
                  function<void(vector<T>&)> sort) {
  Timer timer;
  cout << "Running " << name << "\n";
  ResetStats();
  timer.Tick();
  sort(v);
  timer.Tock();
  ReportStats();
  if (!IsSorted(v)) {
    cout << "sorting algorithm '" << name << "' FAILED!\n";
  }
  cout << "elapsed time:   ";
  cout.width(10);
  cout << std::fixed << std::right << timer.ElapsedTime() << " ms\n";
  cout << "---" << endl;
}

int main(int argc, char * argv[]) {
    vector<int> a{1,2,3,4,5,6,7,8,9,10};
    vector<int> b{1,10,2,9,3,8,4,7,5,6};
    vector<int> c{2,9,4,7,6,5,8,3,10,1};
    vector<int> d{10,9,8,7,6,5,4,3,2,1};
    vector<int> e{1};

    auto sort = Quicksort<int>;

    PrintVector(a);
    sort(a);
    PrintVector(a);
    cout << "---" << endl;
    PrintVector(b);
    sort(b);
    PrintVector(b);
    cout << "---" << endl;
    PrintVector(c);
    sort(c);
    PrintVector(c);
    cout << "---" << endl;
    PrintVector(d);
    sort(d);
    PrintVector(d);
    cout << "---" << endl;
    PrintVector(e);
    sort(e);
    PrintVector(e);
    cout << "---" << endl;

    const size_t size = 10000;
    
    std::cout << "\nsorting a random list\n";
    const auto v = RandomVector<LargeObject>(size);
    // // sort a copy so we get the same stats for all algorithms
    RunAlgorithm<LargeObject>(v, "merge sort", MergeSort<LargeObject>);
    // RunAlgorithm<LargeObject>(v, "bubble sort", BubbleSort<LargeObject>);
    RunAlgorithm<LargeObject>(v, "quicksort", Quicksort<LargeObject>);
    RunAlgorithm<LargeObject>(v, "insertion sort", InsertionSort<LargeObject>);

    // generate a sorted list
    vector<LargeObject> sorted_vector;
    for (unsigned int i = 0; i < size; ++i) {
      sorted_vector.push_back(i);
    }

    std::cout << "\nsorting an already sorted list\n";
    RunAlgorithm<LargeObject>(sorted_vector, "merge sort", MergeSort<LargeObject>);
    // RunAlgorithm<LargeObject>(sorted_vector, "bubble sort", BubbleSort<LargeObject>);
    RunAlgorithm<LargeObject>(sorted_vector, "quicksort", Quicksort<LargeObject>);
    RunAlgorithm<LargeObject>(sorted_vector, "insertion sort", InsertionSort<LargeObject>);

    std::reverse(sorted_vector.begin(), sorted_vector.end());
    std::cout << "\nsorting a list sorted in the opposite order\n";
    RunAlgorithm<LargeObject>(sorted_vector, "merge sort", MergeSort<LargeObject>);
    // RunAlgorithm<LargeObject>(sorted_vector, "bubble sort", BubbleSort<LargeObject>);
    RunAlgorithm<LargeObject>(sorted_vector, "quicksort", Quicksort<LargeObject>);
    RunAlgorithm<LargeObject>(sorted_vector, "insertion sort", InsertionSort<LargeObject>);

    // an interesting edge case: rotated vector
    std::sort(sorted_vector.begin(), sorted_vector.end());
    std::rotate(sorted_vector.begin(), sorted_vector.begin() + 1, sorted_vector.end());
    std::cout << "\nsorting a list that is almost sorted\n";
    RunAlgorithm<LargeObject>(sorted_vector, "merge sort", MergeSort<LargeObject>);
    // RunAlgorithm<LargeObject>(sorted_vector, "bubble sort", BubbleSort<LargeObject>);
    RunAlgorithm<LargeObject>(sorted_vector, "quicksort", Quicksort<LargeObject>);
    RunAlgorithm<LargeObject>(sorted_vector, "insertion sort", InsertionSort<LargeObject>);

    vector<int> w{1, 6, 3};
    cout << *MedianOf3(w.begin(), w.begin() + 1, w.begin() + 2) << endl;

    return 0;
}
